@@ -1,4 +1,4 @@
-/* $Id: upnphttp.c,v 1.99 2014/12/09 17:25:30 nanard Exp $ */
+/* $Id: upnphttp.c,v 1.102 2015/12/12 09:36:22 nanard Exp $ */
 /* Project :  miniupnp
  * Website :  http://miniupnp.free.fr/ or http://miniupnp.tuxfamily.org/
  * Author :   Thomas Bernard
@@ -60,7 +60,7 @@ static int verify_callback(int preverify_ok, X509_STORE_CTX *ctx)
 
 int init_ssl(void)
 {
-	SSL_METHOD *method;
+	const SSL_METHOD *method;
 	SSL_library_init();
 	SSL_load_error_strings();
 	method = TLSv1_server_method();
@@ -665,11 +665,13 @@ with HTTP error 412 Precondition Failed. */
 				BuildResp2_upnphttp(h, 400, "Incompatible header fields", 0, 0);
 			} else
 #endif
-			if(renewSubscription(h->req_buf + h->req_SIDOff, h->req_SIDLen,
-			                     h->req_Timeout) < 0) {
+			sid = upnpevents_renewSubscription(h->req_buf + h->req_SIDOff,
+			                                   h->req_SIDLen, h->req_Timeout);
+			if(!sid) {
 				BuildResp2_upnphttp(h, 412, "Precondition Failed", 0, 0);
 			} else {
-				h->respflags = FLAG_TIMEOUT;
+				h->respflags = FLAG_TIMEOUT | FLAG_SID;
+				h->res_SID = sid;
 				BuildResp_upnphttp(h, 0, 0);
 			}
 		}
@@ -901,7 +903,7 @@ Process_upnphttp(struct upnphttp * h)
 		}
 		else if(n==0)
 		{
-			syslog(LOG_WARNING, "HTTP Connection closed unexpectedly");
+			syslog(LOG_WARNING, "HTTP Connection from %s closed unexpectedly", inet_ntoa(h->clientaddr));
 			h->state = EToDelete;
 		}
 		else
@@ -971,7 +973,7 @@ Process_upnphttp(struct upnphttp * h)
 		}
 		else if(n==0)
 		{
-			syslog(LOG_WARNING, "HTTP Connection closed inexpectedly");
+			syslog(LOG_WARNING, "HTTP Connection from %s closed unexpectedly", inet_ntoa(h->clientaddr));
 			h->state = EToDelete;
 		}
 		else
