@@ -1,7 +1,7 @@
-/* $Id: upnphttp.h,v 1.38 2014/04/09 14:08:13 nanard Exp $ */
+/* $Id: upnphttp.h,v 1.41 2015/09/22 10:07:13 nanard Exp $ */
 /* MiniUPnP project
  * http://miniupnp.free.fr/ or http://miniupnp.tuxfamily.org/
- * (c) 2006-2014 Thomas Bernard
+ * (c) 2006-2015 Thomas Bernard
  * This software is subject to the conditions detailed
  * in the LICENCE file provided within the distribution */
 
@@ -20,10 +20,12 @@
 #if 0
 /* according to "UPnP Device Architecture 1.0" */
 #define UPNP_VERSION_STRING "UPnP/1.0"
-#else
 /* according to "UPnP Device Architecture 1.1" */
 #define UPNP_VERSION_STRING "UPnP/1.1"
+/* according to "UPnP Device Architecture 2.0" */
+#define UPNP_VERSION_STRING "UPnP/2.0"
 #endif
+#define UPNP_VERSION_STRING "UPnP/" UPNP_VERSION_MAJOR_STR "." UPNP_VERSION_MINOR_STR
 
 /* server: HTTP header returned in all HTTP responses : */
 #define MINIUPNPD_SERVER_STRING	OS_VERSION " " UPNP_VERSION_STRING " MiniUPnPd/" MINIUPNPD_VERSION
@@ -72,6 +74,8 @@ struct upnphttp {
 	enum httpCommands req_command;
 	int req_soapActionOff;
 	int req_soapActionLen;
+	int req_HostOff;	/* Host: header */
+	int req_HostLen;
 #ifdef ENABLE_EVENTS
 	int req_CallbackOff;	/* For SUBSCRIBE */
 	int req_CallbackLen;
@@ -136,8 +140,9 @@ Process_upnphttp(struct upnphttp *);
 
 /* BuildHeader_upnphttp()
  * build the header for the HTTP Response
- * also allocate the buffer for body data */
-void
+ * also allocate the buffer for body data
+ * return -1 on error */
+int
 BuildHeader_upnphttp(struct upnphttp * h, int respcode,
                      const char * respmsg,
                      int bodylen);
