@@ -1,7 +1,7 @@
-/* $Id: upnpsoap.c,v 1.127 2014/05/15 10:15:04 nanard Exp $ */
+/* $Id: upnpsoap.c,v 1.140 2015/12/12 09:10:28 nanard Exp $ */
 /* MiniUPnP project
  * http://miniupnp.free.fr/ or http://miniupnp.tuxfamily.org/
- * (c) 2006-2014 Thomas Bernard
+ * (c) 2006-2015 Thomas Bernard
  * This software is subject to the conditions detailed
  * in the LICENCE file provided within the distribution */
 
@@ -45,37 +45,52 @@ BuildSendAndCloseSoapResp(struct upnphttp * h,
 		"</s:Body>"
 		"</s:Envelope>\r\n";
 
-	BuildHeader_upnphttp(h, 200, "OK",  sizeof(beforebody) - 1
-		+ sizeof(afterbody) - 1 + bodylen );
+	int r = BuildHeader_upnphttp(h, 200, "OK",  sizeof(beforebody) - 1
+	                             + sizeof(afterbody) - 1 + bodylen );
 
-	memcpy(h->res_buf + h->res_buflen, beforebody, sizeof(beforebody) - 1);
-	h->res_buflen += sizeof(beforebody) - 1;
+	if(r >= 0) {
+		memcpy(h->res_buf + h->res_buflen, beforebody, sizeof(beforebody) - 1);
+		h->res_buflen += sizeof(beforebody) - 1;
 
-	memcpy(h->res_buf + h->res_buflen, body, bodylen);
-	h->res_buflen += bodylen;
+		memcpy(h->res_buf + h->res_buflen, body, bodylen);
+		h->res_buflen += bodylen;
 
-	memcpy(h->res_buf + h->res_buflen, afterbody, sizeof(afterbody) - 1);
-	h->res_buflen += sizeof(afterbody) - 1;
+		memcpy(h->res_buf + h->res_buflen, afterbody, sizeof(afterbody) - 1);
+		h->res_buflen += sizeof(afterbody) - 1;
+	} else {
+		BuildResp2_upnphttp(h, 500, "Internal Server Error", NULL, 0);
+	}
 
 	SendRespAndClose_upnphttp(h);
 }
 
 static void
-GetConnectionTypeInfo(struct upnphttp * h, const char * action)
+GetConnectionTypeInfo(struct upnphttp * h, const char * action, const char * ns)
 {
+#if 0
 	static const char resp[] =
 		"<u:GetConnectionTypeInfoResponse "
 		"xmlns:u=\"" SERVICE_TYPE_WANIPC "\">"
 		"<NewConnectionType>IP_Routed</NewConnectionType>"
 		"<NewPossibleConnectionTypes>IP_Routed</NewPossibleConnectionTypes>"
 		"</u:GetConnectionTypeInfoResponse>";
-	UNUSED(action);
+#endif
+	static const char resp[] =
+		"<u:%sResponse "
+		"xmlns:u=\"%s\">"
+		"<NewConnectionType>IP_Routed</NewConnectionType>"
+		"<NewPossibleConnectionTypes>IP_Routed</NewPossibleConnectionTypes>"
+		"</u:%sResponse>";
+	char body[512];
+	int bodylen;
 
-	BuildSendAndCloseSoapResp(h, resp, sizeof(resp)-1);
+	bodylen = snprintf(body, sizeof(body), resp,
+	         action, ns, action);
+	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
 
 static void
-GetTotalBytesSent(struct upnphttp * h, const char * action)
+GetTotalBytesSent(struct upnphttp * h, const char * action, const char * ns)
 {
 	int r;
 
@@ -91,13 +106,13 @@ GetTotalBytesSent(struct upnphttp * h, const char * action)
 
 	r = getifstats(ext_if_name, &data);
 	bodylen = snprintf(body, sizeof(body), resp,
-	         action, "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
+	         action, ns, /* was "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1" */
              r<0?0:data.obytes, action);
 	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
 
 static void
-GetTotalBytesReceived(struct upnphttp * h, const char * action)
+GetTotalBytesReceived(struct upnphttp * h, const char * action, const char * ns)
 {
 	int r;
 
@@ -113,13 +128,13 @@ GetTotalBytesReceived(struct upnphttp * h, const char * action)
 
 	r = getifstats(ext_if_name, &data);
 	bodylen = snprintf(body, sizeof(body), resp,
-	         action, "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
+	         action, ns, /* was "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1" */
 	         r<0?0:data.ibytes, action);
 	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
 
 static void
-GetTotalPacketsSent(struct upnphttp * h, const char * action)
+GetTotalPacketsSent(struct upnphttp * h, const char * action, const char * ns)
 {
 	int r;
 
@@ -135,13 +150,13 @@ GetTotalPacketsSent(struct upnphttp * h, const char * action)
 
 	r = getifstats(ext_if_name, &data);
 	bodylen = snprintf(body, sizeof(body), resp,
-	         action, "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
+	         action, ns,/*"urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",*/
 	         r<0?0:data.opackets, action);
 	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
 
 static void
-GetTotalPacketsReceived(struct upnphttp * h, const char * action)
+GetTotalPacketsReceived(struct upnphttp * h, const char * action, const char * ns)
 {
 	int r;
 
@@ -157,21 +172,20 @@ GetTotalPacketsReceived(struct upnphttp * h, const char * action)
 
 	r = getifstats(ext_if_name, &data);
 	bodylen = snprintf(body, sizeof(body), resp,
-	         action, "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
+	         action, ns, /* was "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1" */
 	         r<0?0:data.ipackets, action);
 	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
 
 static void
-GetCommonLinkProperties(struct upnphttp * h, const char * action)
+GetCommonLinkProperties(struct upnphttp * h, const char * action, const char * ns)
 {
 	/* WANAccessType : set depending on the hardware :
 	 * DSL, POTS (plain old Telephone service), Cable, Ethernet */
 	static const char resp[] =
 		"<u:%sResponse "
 		"xmlns:u=\"%s\">"
-		/*"<NewWANAccessType>DSL</NewWANAccessType>"*/
-		"<NewWANAccessType>Cable</NewWANAccessType>"
+		"<NewWANAccessType>%s</NewWANAccessType>"
 		"<NewLayer1UpstreamMaxBitRate>%lu</NewLayer1UpstreamMaxBitRate>"
 		"<NewLayer1DownstreamMaxBitRate>%lu</NewLayer1DownstreamMaxBitRate>"
 		"<NewPhysicalLinkStatus>%s</NewPhysicalLinkStatus>"
@@ -182,6 +196,7 @@ GetCommonLinkProperties(struct upnphttp * h, const char * action)
 	struct ifdata data;
 	const char * status = "Up";	/* Up, Down (Required),
 	                             * Initializing, Unavailable (Optional) */
+	const char * wan_access_type = "Cable"; /* DSL, POTS, Cable, Ethernet */
 	char ext_ip_addr[INET_ADDRSTRLEN];
 
 	if((downstream_bitrate == 0) || (upstream_bitrate == 0))
@@ -196,14 +211,15 @@ GetCommonLinkProperties(struct upnphttp * h, const char * action)
 		status = "Down";
 	}
 	bodylen = snprintf(body, sizeof(body), resp,
-	    action, "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
-		upstream_bitrate, downstream_bitrate,
+	    action, ns, /* was "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1" */
+	    wan_access_type,
+	    upstream_bitrate, downstream_bitrate,
 	    status, action);
 	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
 
 static void
-GetStatusInfo(struct upnphttp * h, const char * action)
+GetStatusInfo(struct upnphttp * h, const char * action, const char * ns)
 {
 	static const char resp[] =
 		"<u:%sResponse "
@@ -224,14 +240,15 @@ GetStatusInfo(struct upnphttp * h, const char * action)
 	status = get_wan_connection_status_str(ext_if_name);
 	uptime = (time(NULL) - startup_time);
 	bodylen = snprintf(body, sizeof(body), resp,
-		action, SERVICE_TYPE_WANIPC,
+		action, ns, /*SERVICE_TYPE_WANIPC,*/
 		status, (long)uptime, action);
 	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
 
 static void
-GetNATRSIPStatus(struct upnphttp * h, const char * action)
+GetNATRSIPStatus(struct upnphttp * h, const char * action, const char * ns)
 {
+#if 0
 	static const char resp[] =
 		"<u:GetNATRSIPStatusResponse "
 		"xmlns:u=\"" SERVICE_TYPE_WANIPC "\">"
@@ -239,6 +256,15 @@ GetNATRSIPStatus(struct upnphttp * h, const char * action)
 		"<NewNATEnabled>1</NewNATEnabled>"
 		"</u:GetNATRSIPStatusResponse>";
 	UNUSED(action);
+#endif
+	static const char resp[] =
+		"<u:%sResponse "
+		"xmlns:u=\"%s\">"
+		"<NewRSIPAvailable>0</NewRSIPAvailable>"
+		"<NewNATEnabled>1</NewNATEnabled>"
+		"</u:%sResponse>";
+	char body[512];
+	int bodylen;
 	/* 2.2.9. RSIPAvailable
 	 * This variable indicates if Realm-specific IP (RSIP) is available
 	 * as a feature on the InternetGatewayDevice. RSIP is being defined
@@ -247,11 +273,14 @@ GetNATRSIPStatus(struct upnphttp * h, const char * action)
 	 * applications that otherwise break if NAT is introduced
 	 * (e.g. IPsec-based VPNs).
 	 * A gateway that does not support RSIP should set this variable to 0. */
-	BuildSendAndCloseSoapResp(h, resp, sizeof(resp)-1);
+	bodylen = snprintf(body, sizeof(body), resp,
+		action, ns, /*SERVICE_TYPE_WANIPC,*/
+		action);
+	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
 
 static void
-GetExternalIPAddress(struct upnphttp * h, const char * action)
+GetExternalIPAddress(struct upnphttp * h, const char * action, const char * ns)
 {
 	static const char resp[] =
 		"<u:%sResponse "
@@ -269,6 +298,7 @@ GetExternalIPAddress(struct upnphttp * h, const char * action)
 	if(use_ext_ip_addr)
 	{
 		strncpy(ext_ip_addr, use_ext_ip_addr, INET_ADDRSTRLEN);
+		ext_ip_addr[INET_ADDRSTRLEN - 1] = '\0';
 	}
 	else if(getifaddr(ext_if_name, ext_ip_addr, INET_ADDRSTRLEN, NULL, NULL) < 0)
 	{
@@ -290,7 +320,7 @@ GetExternalIPAddress(struct upnphttp * h, const char * action)
 	}
 #endif
 	bodylen = snprintf(body, sizeof(body), resp,
-	              action, SERVICE_TYPE_WANIPC,
+	              action, ns, /*SERVICE_TYPE_WANIPC,*/
 				  ext_ip_addr, action);
 	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
@@ -298,14 +328,19 @@ GetExternalIPAddress(struct upnphttp * h, const char * action)
 /* AddPortMapping method of WANIPConnection Service
  * Ignored argument : NewEnabled */
 static void
-AddPortMapping(struct upnphttp * h, const char * action)
+AddPortMapping(struct upnphttp * h, const char * action, const char * ns)
 {
 	int r;
 
-	static const char resp[] =
+	/*static const char resp[] =
 		"<u:AddPortMappingResponse "
-		"xmlns:u=\"" SERVICE_TYPE_WANIPC "\"/>";
+		"xmlns:u=\"" SERVICE_TYPE_WANIPC "\"/>";*/
+	static const char resp[] =
+		"<u:%sResponse "
+		"xmlns:u=\"%s\"/>";
 
+	char body[512];
+	int bodylen;
 	struct NameValueParserData data;
 	char * int_ip, * int_port, * ext_port, * protocol, * desc;
 	char * leaseduration_str;
@@ -445,7 +480,9 @@ AddPortMapping(struct upnphttp * h, const char * action)
 	switch(r)
 	{
 	case 0:	/* success */
-		BuildSendAndCloseSoapResp(h, resp, sizeof(resp)-1);
+		bodylen = snprintf(body, sizeof(body), resp,
+		                   action, ns/*SERVICE_TYPE_WANIPC*/);
+		BuildSendAndCloseSoapResp(h, body, bodylen);
 		break;
 	case -2:	/* already redirected */
 	case -3:	/* not permitted */
@@ -458,7 +495,7 @@ AddPortMapping(struct upnphttp * h, const char * action)
 
 /* AddAnyPortMapping was added in WANIPConnection v2 */
 static void
-AddAnyPortMapping(struct upnphttp * h, const char * action)
+AddAnyPortMapping(struct upnphttp * h, const char * action, const char * ns)
 {
 	int r;
 	static const char resp[] =
@@ -568,7 +605,7 @@ AddAnyPortMapping(struct upnphttp * h, const char * action)
 	{
 	case 0:	/* success */
 		bodylen = snprintf(body, sizeof(body), resp,
-		              action, SERVICE_TYPE_WANIPC,
+		              action, ns, /*SERVICE_TYPE_WANIPC,*/
 					  eport, action);
 		BuildSendAndCloseSoapResp(h, body, bodylen);
 		break;
@@ -584,7 +621,7 @@ AddAnyPortMapping(struct upnphttp * h, const char * action)
 }
 
 static void
-GetSpecificPortMappingEntry(struct upnphttp * h, const char * action)
+GetSpecificPortMappingEntry(struct upnphttp * h, const char * action, const char * ns)
 {
 	int r;
 
@@ -656,7 +693,7 @@ GetSpecificPortMappingEntry(struct upnphttp * h, const char * action)
 		       r_host ? r_host : "NULL", ext_port, protocol, int_ip,
 		       (unsigned int)iport, desc);
 		bodylen = snprintf(body, sizeof(body), resp,
-				action, SERVICE_TYPE_WANIPC,
+				action, ns/*SERVICE_TYPE_WANIPC*/,
 				(unsigned int)iport, int_ip, desc, leaseduration,
 				action);
 		BuildSendAndCloseSoapResp(h, body, bodylen);
@@ -666,29 +703,40 @@ GetSpecificPortMappingEntry(struct upnphttp * h, const char * action)
 }
 
 static void
-DeletePortMapping(struct upnphttp * h, const char * action)
+DeletePortMapping(struct upnphttp * h, const char * action, const char * ns)
 {
 	int r;
 
-	static const char resp[] =
+	/*static const char resp[] =
 		"<u:DeletePortMappingResponse "
 		"xmlns:u=\"" SERVICE_TYPE_WANIPC "\">"
-		"</u:DeletePortMappingResponse>";
+		"</u:DeletePortMappingResponse>";*/
+	static const char resp[] =
+		"<u:%sResponse "
+		"xmlns:u=\"%s\">"
+		"</u:%sResponse>";
 
+	char body[512];
+	int bodylen;
 	struct NameValueParserData data;
-	const char * r_host, * ext_port, * protocol;
+	const char * ext_port, * protocol;
 	unsigned short eport;
+#ifdef UPNP_STRICT
+	const char * r_host;
+#endif /* UPNP_STRICT */
 
 	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
-	r_host = GetValueFromNameValueList(&data, "NewRemoteHost");
 	ext_port = GetValueFromNameValueList(&data, "NewExternalPort");
 	protocol = GetValueFromNameValueList(&data, "NewProtocol");
+#ifdef UPNP_STRICT
+	r_host = GetValueFromNameValueList(&data, "NewRemoteHost");
+#endif /* UPNP_STRICT */
 
 #ifdef UPNP_STRICT
 	if(!ext_port || !protocol || !r_host)
 #else
 	if(!ext_port || !protocol)
-#endif
+#endif /* UPNP_STRICT */
 	{
 		ClearNameValueList(&data);
 		SoapError(h, 402, "Invalid Args");
@@ -702,18 +750,42 @@ DeletePortMapping(struct upnphttp * h, const char * action)
 		SoapError(h, 726, "RemoteHostOnlySupportsWildcard");
 		return;
 	}
-#endif
-#endif
+#endif /* UPNP_STRICT */
+#endif /* SUPPORT_REMOTEHOST */
 
 	eport = (unsigned short)atoi(ext_port);
 
-	/* TODO : if in secure mode, check the IP
+	syslog(LOG_INFO, "%s: external port: %hu, protocol: %s",
+		action, eport, protocol);
+
+	/* if in secure mode, check the IP
 	 * Removing a redirection is not a security threat,
 	 * just an annoyance for the user using it. So this is not
 	 * a priority. */
-
-	syslog(LOG_INFO, "%s: external port: %hu, protocol: %s",
-		action, eport, protocol);
+	if(GETFLAG(SECUREMODEMASK))
+	{
+		char int_ip[32];
+		struct in_addr int_ip_addr;
+		unsigned short iport;
+		unsigned int leaseduration = 0;
+		r = upnp_get_redirection_infos(eport, protocol, &iport,
+		                               int_ip, sizeof(int_ip),
+		                               NULL, 0, NULL, 0,
+		                               &leaseduration);
+		if(r >= 0)
+		{
+			if(inet_pton(AF_INET, int_ip, &int_ip_addr) > 0)
+			{
+				if(h->clientaddr.s_addr != int_ip_addr.s_addr)
+				{
+					SoapError(h, 606, "Action not authorized");
+					/*SoapError(h, 714, "NoSuchEntryInArray");*/
+					ClearNameValueList(&data);
+					return;
+				}
+			}
+		}
+	}
 
 	r = upnp_delete_redirection(eport, protocol);
 
@@ -723,7 +795,9 @@ DeletePortMapping(struct upnphttp * h, const char * action)
 	}
 	else
 	{
-		BuildSendAndCloseSoapResp(h, resp, sizeof(resp)-1);
+		bodylen = snprintf(body, sizeof(body), resp,
+		                   action, ns, action);
+		BuildSendAndCloseSoapResp(h, body, bodylen);
 	}
 
 	ClearNameValueList(&data);
@@ -731,13 +805,19 @@ DeletePortMapping(struct upnphttp * h, const char * action)
 
 /* DeletePortMappingRange was added in IGD spec v2 */
 static void
-DeletePortMappingRange(struct upnphttp * h, const char * action)
+DeletePortMappingRange(struct upnphttp * h, const char * action, const char * ns)
 {
 	int r = -1;
-	static const char resp[] =
+	/*static const char resp[] =
 		"<u:DeletePortMappingRangeResponse "
 		"xmlns:u=\"" SERVICE_TYPE_WANIPC "\">"
-		"</u:DeletePortMappingRangeResponse>";
+		"</u:DeletePortMappingRangeResponse>";*/
+	static const char resp[] =
+		"<u:%sResponse "
+		"xmlns:u=\"%s\">"
+		"</u:%sResponse>";
+	char body[512];
+	int bodylen;
 	struct NameValueParserData data;
 	const char * protocol;
 	const char * startport_s, * endport_s;
@@ -745,7 +825,6 @@ DeletePortMappingRange(struct upnphttp * h, const char * action)
 	/*int manage;*/
 	unsigned short * port_list;
 	unsigned int i, number = 0;
-	UNUSED(action);
 
 	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
 	startport_s = GetValueFromNameValueList(&data, "NewStartPort");
@@ -781,6 +860,7 @@ DeletePortMappingRange(struct upnphttp * h, const char * action)
 	{
 		SoapError(h, 730, "PortMappingNotFound");
 		ClearNameValueList(&data);
+		free(port_list);
 		return;
 	}
 
@@ -791,13 +871,15 @@ DeletePortMappingRange(struct upnphttp * h, const char * action)
 		       action, port_list[i], protocol, r < 0 ? "failed" : "ok");
 	}
 	free(port_list);
-	BuildSendAndCloseSoapResp(h, resp, sizeof(resp)-1);
+	bodylen = snprintf(body, sizeof(body), resp,
+	                   action, ns, action);
+	BuildSendAndCloseSoapResp(h, body, bodylen);
 
 	ClearNameValueList(&data);
 }
 
 static void
-GetGenericPortMappingEntry(struct upnphttp * h, const char * action)
+GetGenericPortMappingEntry(struct upnphttp * h, const char * action, const char * ns)
 {
 	int r;
 
@@ -868,7 +950,7 @@ GetGenericPortMappingEntry(struct upnphttp * h, const char * action)
 		int bodylen;
 		char body[2048];
 		bodylen = snprintf(body, sizeof(body), resp,
-			action, SERVICE_TYPE_WANIPC, rhost,
+			action, ns, /*SERVICE_TYPE_WANIPC,*/ rhost,
 			(unsigned int)eport, protocol, (unsigned int)iport, iaddr, desc,
 		    leaseduration, action);
 		BuildSendAndCloseSoapResp(h, body, bodylen);
@@ -879,7 +961,7 @@ GetGenericPortMappingEntry(struct upnphttp * h, const char * action)
 
 /* GetListOfPortMappings was added in the IGD v2 specification */
 static void
-GetListOfPortMappings(struct upnphttp * h, const char * action)
+GetListOfPortMappings(struct upnphttp * h, const char * action, const char * ns)
 {
 	static const char resp_start[] =
 		"<u:%sResponse "
@@ -983,7 +1065,7 @@ http://www.upnp.org/schemas/gw/WANIPConnection-v2.xsd">
 		return;
 	}
 	bodylen = snprintf(body, bodyalloc, resp_start,
-	              action, SERVICE_TYPE_WANIPC);
+	              action, ns/*SERVICE_TYPE_WANIPC*/);
 	if(bodylen < 0)
 	{
 		SoapError(h, 501, "ActionFailed");
@@ -1006,6 +1088,7 @@ http://www.upnp.org/schemas/gw/WANIPConnection-v2.xsd">
 			body = realloc(body, bodyalloc);
 			if(!body)
 			{
+				syslog(LOG_CRIT, "realloc(%p, %u) FAILED", body_sav, (unsigned)bodyalloc);
 				ClearNameValueList(&data);
 				SoapError(h, 501, "ActionFailed");
 				free(body_sav);
@@ -1030,6 +1113,20 @@ http://www.upnp.org/schemas/gw/WANIPConnection-v2.xsd">
 	free(port_list);
 	port_list = NULL;
 
+	if((bodylen + sizeof(list_end) + 1024) > bodyalloc)
+	{
+		char * body_sav = body;
+		bodyalloc += (sizeof(list_end) + 1024);
+		body = realloc(body, bodyalloc);
+		if(!body)
+		{
+			syslog(LOG_CRIT, "realloc(%p, %u) FAILED", body_sav, (unsigned)bodyalloc);
+			ClearNameValueList(&data);
+			SoapError(h, 501, "ActionFailed");
+			free(body_sav);
+			return;
+		}
+	}
 	memcpy(body+bodylen, list_end, sizeof(list_end));
 	bodylen += (sizeof(list_end) - 1);
 	bodylen += snprintf(body+bodylen, bodyalloc-bodylen, resp_end,
@@ -1042,12 +1139,18 @@ http://www.upnp.org/schemas/gw/WANIPConnection-v2.xsd">
 
 #ifdef ENABLE_L3F_SERVICE
 static void
-SetDefaultConnectionService(struct upnphttp * h, const char * action)
+SetDefaultConnectionService(struct upnphttp * h, const char * action, const char * ns)
 {
-	static const char resp[] =
+	/*static const char resp[] =
 		"<u:SetDefaultConnectionServiceResponse "
 		"xmlns:u=\"urn:schemas-upnp-org:service:Layer3Forwarding:1\">"
-		"</u:SetDefaultConnectionServiceResponse>";
+		"</u:SetDefaultConnectionServiceResponse>";*/
+	static const char resp[] =
+		"<u:%sResponse "
+		"xmlns:u=\"%s\">"
+		"</u:%sResponse>";
+	char body[512];
+	int bodylen;
 	struct NameValueParserData data;
 	char * p;
 	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
@@ -1067,7 +1170,9 @@ SetDefaultConnectionService(struct upnphttp * h, const char * action)
 #endif
 		{
 			syslog(LOG_INFO, "%s(%s) : Ignored", action, p);
-			BuildSendAndCloseSoapResp(h, resp, sizeof(resp)-1);
+			bodylen = snprintf(body, sizeof(body), resp,
+			                   action, ns, action);
+			BuildSendAndCloseSoapResp(h, body, bodylen);
 		}
 	} else {
 		/* missing argument */
@@ -1077,12 +1182,16 @@ SetDefaultConnectionService(struct upnphttp * h, const char * action)
 }
 
 static void
-GetDefaultConnectionService(struct upnphttp * h, const char * action)
+GetDefaultConnectionService(struct upnphttp * h, const char * action, const char * ns)
 {
 	static const char resp[] =
 		"<u:%sResponse "
-		"xmlns:u=\"urn:schemas-upnp-org:service:Layer3Forwarding:1\">"
+		"xmlns:u=\"%s\">"
+#ifdef IGD_V2
+		"<NewDefaultConnectionService>%s:WANConnectionDevice:2,"
+#else
 		"<NewDefaultConnectionService>%s:WANConnectionDevice:1,"
+#endif
 		SERVICE_ID_WANIPC "</NewDefaultConnectionService>"
 		"</u:%sResponse>";
 	/* example from UPnP_IGD_Layer3Forwarding 1.0.pdf :
@@ -1091,29 +1200,33 @@ GetDefaultConnectionService(struct upnphttp * h, const char * action)
 	char body[1024];
 	int bodylen;
 
+	/* namespace : urn:schemas-upnp-org:service:Layer3Forwarding:1 */
 	bodylen = snprintf(body, sizeof(body), resp,
-	                   action, uuidvalue_wcd, action);
+	                   action, ns, uuidvalue_wcd, action);
 	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
 #endif
 
 /* Added for compliance with WANIPConnection v2 */
 static void
-SetConnectionType(struct upnphttp * h, const char * action)
+SetConnectionType(struct upnphttp * h, const char * action, const char * ns)
 {
+#ifdef UPNP_STRICT
 	const char * connection_type;
+#endif /* UPNP_STRICT */
 	struct NameValueParserData data;
 	UNUSED(action);
+	UNUSED(ns);
 
 	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
-	connection_type = GetValueFromNameValueList(&data, "NewConnectionType");
 #ifdef UPNP_STRICT
+	connection_type = GetValueFromNameValueList(&data, "NewConnectionType");
 	if(!connection_type) {
 		ClearNameValueList(&data);
 		SoapError(h, 402, "Invalid Args");
 		return;
 	}
-#endif
+#endif /* UPNP_STRICT */
 	/* Unconfigured, IP_Routed, IP_Bridged */
 	ClearNameValueList(&data);
 	/* always return a ReadOnly error */
@@ -1122,17 +1235,19 @@ SetConnectionType(struct upnphttp * h, const char * action)
 
 /* Added for compliance with WANIPConnection v2 */
 static void
-RequestConnection(struct upnphttp * h, const char * action)
+RequestConnection(struct upnphttp * h, const char * action, const char * ns)
 {
 	UNUSED(action);
+	UNUSED(ns);
 	SoapError(h, 606, "Action not authorized");
 }
 
 /* Added for compliance with WANIPConnection v2 */
 static void
-ForceTermination(struct upnphttp * h, const char * action)
+ForceTermination(struct upnphttp * h, const char * action, const char * ns)
 {
 	UNUSED(action);
+	UNUSED(ns);
 	SoapError(h, 606, "Action not authorized");
 }
 
@@ -1145,7 +1260,7 @@ QueryStateVariable remains useful as a limited test tool but may not be
 part of some future versions of UPnP.
 */
 static void
-QueryStateVariable(struct upnphttp * h, const char * action)
+QueryStateVariable(struct upnphttp * h, const char * action, const char * ns)
 {
 	static const char resp[] =
         "<u:%sResponse "
@@ -1175,7 +1290,7 @@ QueryStateVariable(struct upnphttp * h, const char * action)
 
 		status = get_wan_connection_status_str(ext_if_name);
 		bodylen = snprintf(body, sizeof(body), resp,
-                           action, "urn:schemas-upnp-org:control-1-0",
+                           action, ns,/*"urn:schemas-upnp-org:control-1-0",*/
 		                   status, action);
 		BuildSendAndCloseSoapResp(h, body, bodylen);
 	}
@@ -1198,7 +1313,7 @@ QueryStateVariable(struct upnphttp * h, const char * action)
 		snprintf(strn, sizeof(strn), "%i",
 		         upnp_get_portmapping_number_of_entries());
 		bodylen = snprintf(body, sizeof(body), resp,
-                           action, "urn:schemas-upnp-org:control-1-0",
+                           action, ns,/*"urn:schemas-upnp-org:control-1-0",*/
 		                   strn, action);
 		BuildSendAndCloseSoapResp(h, body, bodylen);
 	}
@@ -1217,7 +1332,7 @@ QueryStateVariable(struct upnphttp * h, const char * action)
 #endif
 /* WANIPv6FirewallControl actions */
 static void
-GetFirewallStatus(struct upnphttp * h, const char * action)
+GetFirewallStatus(struct upnphttp * h, const char * action, const char * ns)
 {
 	static const char resp[] =
 		"<u:%sResponse "
@@ -1230,7 +1345,7 @@ GetFirewallStatus(struct upnphttp * h, const char * action)
 	int bodylen;
 
 	bodylen = snprintf(body, sizeof(body), resp,
-		action, "urn:schemas-upnp-org:service:WANIPv6FirewallControl:1",
+		action, ns, /*"urn:schemas-upnp-org:service:WANIPv6FirewallControl:1",*/
 	    GETFLAG(IPV6FCFWDISABLEDMASK) ? 0 : 1,
 	    GETFLAG(IPV6FCINBOUNDDISALLOWEDMASK) ? 0 : 1,
 	    action);
@@ -1374,7 +1489,7 @@ PinholeVerification(struct upnphttp * h, char * int_ip, unsigned short int_port)
 }
 
 static void
-AddPinhole(struct upnphttp * h, const char * action)
+AddPinhole(struct upnphttp * h, const char * action, const char * ns)
 {
 	int r;
 	static const char resp[] =
@@ -1433,7 +1548,7 @@ AddPinhole(struct upnphttp * h, const char * action)
 	}
 	/* I guess it is useless to convert int_ip to literal ipv6 address */
 	/* rem_host should be converted to literal ipv6 : */
-	if(rem_host)
+	if(rem_host && (rem_host[0] != '\0'))
 	{
 		struct addrinfo *ai, *p;
 		struct addrinfo hints;
@@ -1512,7 +1627,7 @@ AddPinhole(struct upnphttp * h, const char * action)
 		case 1:	        /* success */
 			bodylen = snprintf(body, sizeof(body),
 			                   resp, action,
-			                   "urn:schemas-upnp-org:service:WANIPv6FirewallControl:1",
+			                   ns/*"urn:schemas-upnp-org:service:WANIPv6FirewallControl:1"*/,
 			                   uid, action);
 			BuildSendAndCloseSoapResp(h, body, bodylen);
 			break;
@@ -1536,12 +1651,20 @@ clear_and_exit:
 }
 
 static void
-UpdatePinhole(struct upnphttp * h, const char * action)
+UpdatePinhole(struct upnphttp * h, const char * action, const char * ns)
 {
+#if 0
 	static const char resp[] =
 		"<u:UpdatePinholeResponse "
 		"xmlns:u=\"urn:schemas-upnp-org:service:WANIPv6FirewallControl:1\">"
 		"</u:UpdatePinholeResponse>";
+#endif
+	static const char resp[] =
+		"<u:%sResponse "
+		"xmlns:u=\"%s\">"
+		"</u:%sResponse>";
+	char body[512];
+	int bodylen;
 	struct NameValueParserData data;
 	const char * uid_str, * leaseTime;
 	char iaddr[INET6_ADDRSTRLEN];
@@ -1597,12 +1720,15 @@ UpdatePinhole(struct upnphttp * h, const char * action)
 		SoapError(h, 704, "NoSuchEntry");
 	else if(n < 0)
 		SoapError(h, 501, "ActionFailed");
-	else
-		BuildSendAndCloseSoapResp(h, resp, sizeof(resp)-1);
+	else {
+		bodylen = snprintf(body, sizeof(body), resp,
+		                   action, ns, action);
+		BuildSendAndCloseSoapResp(h, body, bodylen);
+	}
 }
 
 static void
-GetOutboundPinholeTimeout(struct upnphttp * h, const char * action)
+GetOutboundPinholeTimeout(struct upnphttp * h, const char * action, const char * ns)
 {
 	int r;
 
@@ -1616,7 +1742,8 @@ GetOutboundPinholeTimeout(struct upnphttp * h, const char * action)
 	int bodylen;
 	struct NameValueParserData data;
 	char * int_ip, * int_port, * rem_host, * rem_port, * protocol;
-	int opt=0, proto=0;
+	int opt=0;
+	/*int proto=0;*/
 	unsigned short iport, rport;
 
 	if (GETFLAG(IPV6FCFWDISABLEDMASK))
@@ -1634,7 +1761,7 @@ GetOutboundPinholeTimeout(struct upnphttp * h, const char * action)
 
 	rport = (unsigned short)atoi(rem_port);
 	iport = (unsigned short)atoi(int_port);
-	proto = atoi(protocol);
+	/*proto = atoi(protocol);*/
 
 	syslog(LOG_INFO, "%s: retrieving timeout for outbound pinhole from [%s]:%hu to [%s]:%hu protocol %s", action, int_ip, iport,rem_host, rport, protocol);
 
@@ -1645,7 +1772,7 @@ GetOutboundPinholeTimeout(struct upnphttp * h, const char * action)
 	{
 		case 1:	/* success */
 			bodylen = snprintf(body, sizeof(body), resp,
-			                   action, "urn:schemas-upnp-org:service:WANIPv6FirewallControl:1",
+			                   action, ns/*"urn:schemas-upnp-org:service:WANIPv6FirewallControl:1"*/,
 			                   opt, action);
 			BuildSendAndCloseSoapResp(h, body, bodylen);
 			break;
@@ -1659,14 +1786,21 @@ GetOutboundPinholeTimeout(struct upnphttp * h, const char * action)
 }
 
 static void
-DeletePinhole(struct upnphttp * h, const char * action)
+DeletePinhole(struct upnphttp * h, const char * action, const char * ns)
 {
 	int n;
-
+#if 0
 	static const char resp[] =
 		"<u:DeletePinholeResponse "
 		"xmlns:u=\"urn:schemas-upnp-org:service:WANIPv6FirewallControl:1\">"
 		"</u:DeletePinholeResponse>";
+#endif
+	static const char resp[] =
+		"<u:%sResponse "
+		"xmlns:u=\"%s\">"
+		"</u:%sResponse>";
+	char body[512];
+	int bodylen;
 
 	struct NameValueParserData data;
 	const char * uid_str;
@@ -1723,11 +1857,13 @@ DeletePinhole(struct upnphttp * h, const char * action)
 	}
 	syslog(LOG_INFO, "%s: (inbound) pinhole with ID %d successfully removed",
 	       action, uid);
-	BuildSendAndCloseSoapResp(h, resp, sizeof(resp)-1);
+	bodylen = snprintf(body, sizeof(body), resp,
+	                   action, ns, action);
+	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
 
 static void
-CheckPinholeWorking(struct upnphttp * h, const char * action)
+CheckPinholeWorking(struct upnphttp * h, const char * action, const char * ns)
 {
 	static const char resp[] =
 		"<u:%sResponse "
@@ -1776,7 +1912,7 @@ CheckPinholeWorking(struct upnphttp * h, const char * action)
 			return;
 		}
 		bodylen = snprintf(body, sizeof(body), resp,
-						action, "urn:schemas-upnp-org:service:WANIPv6FirewallControl:1",
+						action, ns/*"urn:schemas-upnp-org:service:WANIPv6FirewallControl:1"*/,
 						1, action);
 		BuildSendAndCloseSoapResp(h, body, bodylen);
 	}
@@ -1787,7 +1923,7 @@ CheckPinholeWorking(struct upnphttp * h, const char * action)
 }
 
 static void
-GetPinholePackets(struct upnphttp * h, const char * action)
+GetPinholePackets(struct upnphttp * h, const char * action, const char * ns)
 {
 	static const char resp[] =
 		"<u:%sResponse "
@@ -1840,7 +1976,7 @@ GetPinholePackets(struct upnphttp * h, const char * action)
 #endif
 
 	bodylen = snprintf(body, sizeof(body), resp,
-			action, "urn:schemas-upnp-org:service:WANIPv6FirewallControl:1",
+			action, ns/*"urn:schemas-upnp-org:service:WANIPv6FirewallControl:1"*/,
 			packets, action);
 	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
@@ -1848,12 +1984,12 @@ GetPinholePackets(struct upnphttp * h, const char * action)
 
 #ifdef ENABLE_DP_SERVICE
 static void
-SendSetupMessage(struct upnphttp * h, const char * action)
+SendSetupMessage(struct upnphttp * h, const char * action, const char * ns)
 {
 	static const char resp[] =
 		"<u:%sResponse "
 		"xmlns:u=\"%s\">"
-		"<NewOutMessage>%s</NewOutMessage>"
+		"<OutMessage>%s</OutMessage>"
 		"</u:%sResponse>";
 	char body[1024];
 	int bodylen;
@@ -1882,19 +2018,19 @@ SendSetupMessage(struct upnphttp * h, const char * action)
 	/* TODO : put here code for WPS */
 
 	bodylen = snprintf(body, sizeof(body), resp,
-	                   action, "urn:schemas-upnp-org:service:DeviceProtection:1",
+	                   action, ns/*"urn:schemas-upnp-org:service:DeviceProtection:1"*/,
 	                   OutMessage, action);
 	BuildSendAndCloseSoapResp(h, body, bodylen);
 	ClearNameValueList(&data);
 }
 
 static void
-GetSupportedProtocols(struct upnphttp * h, const char * action)
+GetSupportedProtocols(struct upnphttp * h, const char * action, const char * ns)
 {
 	static const char resp[] =
 		"<u:%sResponse "
 		"xmlns:u=\"%s\">"
-		"<NewProtocolList>%s</NewProtocolList>"
+		"<ProtocolList>%s</ProtocolList>"
 		"</u:%sResponse>";
 	char body[1024];
 	int bodylen;
@@ -1909,18 +2045,18 @@ GetSupportedProtocols(struct upnphttp * h, const char * action)
 		"</SupportedProtocols>";
 
 	bodylen = snprintf(body, sizeof(body), resp,
-	                   action, "urn:schemas-upnp-org:service:DeviceProtection:1",
+	                   action, ns/*"urn:schemas-upnp-org:service:DeviceProtection:1"*/,
 	                   ProtocolList, action);
 	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
 
 static void
-GetAssignedRoles(struct upnphttp * h, const char * action)
+GetAssignedRoles(struct upnphttp * h, const char * action, const char * ns)
 {
 	static const char resp[] =
 		"<u:%sResponse "
 		"xmlns:u=\"%s\">"
-		"<NewRoleList>%s</NewRoleList>"
+		"<RoleList>%s</RoleList>"
 		"</u:%sResponse>";
 	char body[1024];
 	int bodylen;
@@ -1939,7 +2075,7 @@ GetAssignedRoles(struct upnphttp * h, const char * action)
 #endif
 
 	bodylen = snprintf(body, sizeof(body), resp,
-	                   action, "urn:schemas-upnp-org:service:DeviceProtection:1",
+	                   action, ns/*"urn:schemas-upnp-org:service:DeviceProtection:1"*/,
 	                   RoleList, action);
 	BuildSendAndCloseSoapResp(h, body, bodylen);
 }
@@ -1960,7 +2096,7 @@ GetAssignedRoles(struct upnphttp * h, const char * action)
 static const struct
 {
 	const char * methodName;
-	void (*methodImpl)(struct upnphttp *, const char *);
+	void (*methodImpl)(struct upnphttp *, const char *, const char *);
 }
 soapMethods[] =
 {
@@ -2017,35 +2153,37 @@ ExecuteSoapAction(struct upnphttp * h, const char * action, int n)
 	char * p;
 	char * p2;
 	int i, len, methodlen;
+	char namespace[256];
 
-	i = 0;
+	/* SoapAction example :
+	 * urn:schemas-upnp-org:service:WANIPConnection:1#GetStatusInfo */
 	p = strchr(action, '#');
-
-	if(p)
-	{
+	if(p && (p - action) < n) {
+		for(i = 0; i < ((int)sizeof(namespace) - 1) && (action + i) < p; i++)
+			namespace[i] = action[i];
+		namespace[i] = '\0';
 		p++;
 		p2 = strchr(p, '"');
-		if(p2)
+		if(p2 && (p2 - action) <= n)
 			methodlen = p2 - p;
 		else
 			methodlen = n - (p - action);
-		/*syslog(LOG_DEBUG, "SoapMethod: %.*s", methodlen, p);*/
-		while(soapMethods[i].methodName)
-		{
+		/*syslog(LOG_DEBUG, "SoapMethod: %.*s %d %d %p %p %d",
+		       methodlen, p, methodlen, n, action, p, (int)(p - action));*/
+		for(i = 0; soapMethods[i].methodName; i++) {
 			len = strlen(soapMethods[i].methodName);
-			if((len == methodlen) && memcmp(p, soapMethods[i].methodName, len) == 0)
-			{
+			if((len == methodlen) && memcmp(p, soapMethods[i].methodName, len) == 0) {
 #ifdef DEBUG
-				syslog(LOG_DEBUG, "Remote Call of SoapMethod '%s'\n",
+				syslog(LOG_DEBUG, "Remote Call of SoapMethod '%s'",
 				       soapMethods[i].methodName);
-#endif
-				soapMethods[i].methodImpl(h, soapMethods[i].methodName);
+#endif /* DEBUG */
+				soapMethods[i].methodImpl(h, soapMethods[i].methodName, namespace);
 				return;
 			}
-			i++;
 		}
-
 		syslog(LOG_NOTICE, "SoapMethod: Unknown: %.*s", methodlen, p);
+	} else {
+		syslog(LOG_NOTICE, "cannot parse SoapAction");
 	}
 
 	SoapError(h, 401, "Invalid Action");
